#!/bin/bash
#
# gmsh <=> dolfyn script
#
# script still assumes that dolfyn's geometry is in
# a file with extension *.geo. as of dolfyn 0.5.27
# this file uses the extension *.dge in order not to
# collide with Gmsh. all three dolfyn codes use *.dge
# by default and use *.geo when a *.dge does not exist.
#
# set paths to the codes 
# (when not in path just absolute or relative):
#
gmsh=`which gmsh`

gmsh2dolfyn="../bin/gmsh2dolfyn"
preprocessor="../bin/doldge"
dolfyn="../bin/dolfyn"

#
# the tutorials
# (set to 1 to activate)
#
TUT1=
TUT2=
TUT3=
TUT4=
TUT5=1
TUT6=

if test $TUT1 ; then

  echo First model

  MDL=t2
  rm ${MDL}.msh
  rm ${MDL}d.cel ${MDL}d.vrt ${MDL}d.bnd ${MDL}d.mon 
  rm ${MDL}d.res ${MDL}d.rst ${MDL}d.geo ${MDL}d.inp
  rm ${MDL}d.msh ${MDL}d_??.msh  

  $gmsh -3 -optimize t2.geo

  # copy!
  cp t2.msh t2d.msh

  echo "t2d"                | $gmsh2dolfyn 
  echo -e "t2d\n1.0\nbin\n" | $preprocessor
  echo "t2d"                | $dolfyn

  $gmsh t2.msh t2d_pv.msh t2d_uvwc.msh

fi

if test $TUT2 ; then

  echo Second model

  MDL=t3a
  rm ${MDL}.msh
  rm ${MDL}d.cel ${MDL}d.vrt ${MDL}d.bnd ${MDL}d.mon 
  rm ${MDL}d.res ${MDL}d.rst ${MDL}d.geo ${MDL}d.inp
  rm ${MDL}d.msh ${MDL}d_??.msh  

  $gmsh -3 -optimize t3a.geo

  # copy!
  cp t3a.msh t3ad.msh

  echo "t3ad"                | $gmsh2dolfyn 
  echo -e "t3ad\n1.0\nbin\n" | $preprocessor
  echo "t3ad"                | $dolfyn

  $gmsh t3ad.msh t3ad_uvwc.msh

fi

if test $TUT3 ; then

  echo Second model revisited

  MDL=t3b
  rm ${MDL}.msh
  rm ${MDL}d.cel ${MDL}d.vrt ${MDL}d.bnd ${MDL}d.mon 
  rm ${MDL}d.res ${MDL}d.rst ${MDL}d.geo ${MDL}d.inp
  rm ${MDL}d.msh ${MDL}d_??.msh  

  $gmsh -3 -optimize t3b.geo

  # copy!
  cp t3b.msh t3bd.msh

  echo "t3bd"                | $gmsh2dolfyn 
  echo -e "t3bd\n1.0\nbin\n" | $preprocessor
  echo "t3bd"                | $dolfyn

  $gmsh t3bd.msh t3bd_uvwc.msh

fi

if test $TUT4 ; then

  echo Fourth model

  MDL=t4
  rm ${MDL}.msh
  rm ${MDL}d.cel ${MDL}d.vrt ${MDL}d.bnd ${MDL}d.mon 
  rm ${MDL}d.res ${MDL}d.rst ${MDL}d.geo ${MDL}d.inp
  rm ${MDL}d.msh ${MDL}d_??.msh  

  $gmsh -3 -optimize t4.geo

  # copy!
  cp t4.msh t4d.msh

  echo "t4d"                | $gmsh2dolfyn 
  echo -e "t4d\n1.0\nbin\n" | $preprocessor
  echo "t4d"                | $dolfyn

  $gmsh t4d.msh t4d_uvwc.msh

fi

if test $TUT5 ; then

  echo Fifth model

  MDL=t5
  rm ${MDL}.msh
  rm ${MDL}d.cel ${MDL}d.vrt ${MDL}d.bnd ${MDL}d.mon 
  rm ${MDL}d.res ${MDL}d.rst ${MDL}d.geo ${MDL}d.inp
  rm ${MDL}d.msh ${MDL}d_??.msh  

  $gmsh -3 -optimize t5.geo

  # copy!
  cp t5.msh t5d.msh

  echo "t5d"                | $gmsh2dolfyn 
  echo -e "t5d\n1.0\nbin\n" | $preprocessor
  echo "t5d"                | $dolfyn

  $gmsh t5d.msh t5d_tc.msh

fi

if test $TUT6 ; then

  echo Sixth model

  MDL=t6
  rm ${MDL}.msh
  rm ${MDL}d.cel ${MDL}d.vrt ${MDL}d.bnd ${MDL}d.mon 
  rm ${MDL}d.res ${MDL}d.rst ${MDL}d.geo ${MDL}d.inp
  rm ${MDL}d.msh ${MDL}d_??.msh  

  $gmsh -3 -optimize t6.geo

  # copy!
  cp t6.msh t6d.msh

  echo "t6d"                | $gmsh2dolfyn 
  echo -e "t6d\n1.0\nbin\n" | $preprocessor
  echo "t6d"                | $dolfyn

  $gmsh t6.msh t6d_uvwc.msh t6d_pc.msh

fi

echo Done
